from __future__ import annotations

from typing import Union, List

from treepath.path.traverser.traverser_match import TraverserMatch
from treepath.path.vertex.vertex import Vertex


class Match:
    """
    The Match Class contains all the details that describe the Match.
    """
    __slots__ = '_traverser_match'

    def __init__(self,
                 traverser_match: TraverserMatch,
                 ):
        self._traverser_match = traverser_match

    @property
    def path_as_list(self) -> List[Match]:
        """
        Returns a list containing this Mtch and all the ancestor matches.  The list is order from root to this Match.
        """
        return [Match(traverser_match) for traverser_match in self._traverser_match.path_as_list]

    @property
    def path(self) -> str:
        """
        Returns the str representation of the absolute path for this match.  For example $.a.b[0]
        """
        return self._traverser_match.path

    @property
    def path_segment(self):
        """
        Returns the str representation of the segment in the path this Match represents.   This is different from
        data_name in that it includes additional symbols.  For example if data_name is a list index of '1' then path
        segment is '[1]'
        """
        return self._traverser_match.path_segment

    @property
    def parent(self) -> Union[Match, None]:
        """
        Returns the parent Match.
        """
        parent = self._traverser_match.parent
        if parent:
            return Match(parent)
        else:
            return None

    @property
    def data_name(self) -> Union[str, int]:
        """
        Return the dictionary key or the list index this Match represents.
        """
        return self._traverser_match.data_name

    @property
    def data(self) -> Union[dict, list, str, int, float, bool, None]:
        """
        Return the value the data_name references.
        """
        return self._traverser_match.data

    @property
    def vertex(self) -> Vertex:
        """
        Returns the path vertex that generated this match.
        """
        return self._traverser_match.vertex

    def __repr__(self):
        """
        Return  the match in the format f"{self.path}=={self.data}"
        """
        return repr(self._traverser_match)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        """
        Return True if other Match data_name, data and parent are all equal.
        """
        return (
                self is other
                or (
                        isinstance(other, Match)
                        and (
                                self._traverser_match is other._traverser_match
                                or (
                                        self.data == other.data
                                        and self.data_name == other.data_name
                                        and self.parent == other.parent
                                )
                        )
                )
        )

    def __ne__(self, other):
        return not self.__eq__(other)



from typing import Union

from treepath.path.traverser.imaginary_match import ImaginaryMatch
from treepath.path.traverser.key_match import KeyMatch
from treepath.path.traverser.list_match import ListMatch
from treepath.path.traverser.traverser_match import TraverserMatch
from treepath.path.vertex.vertex import Vertex


class CatchState:
    __slots__ = 'iterable', \
                'match_constructor'

    def __init__(self, iterable, match_constructor):
        self.iterable = iterable
        self.match_constructor = match_constructor


class RecursiveVertex(Vertex):
    __slots__ = ()

    def __init__(self, parent):
        super().__init__(parent, ".")
        self.is_catch_vertex = True

    @property
    def path_segment(self):
        return f"."

    @property
    def path(self):
        return ''.join(vertex.path_segment for vertex in self.path_as_list) + "."

    def match(self, parent_match: TraverserMatch, traverser, vertex_index: int) -> Union[TraverserMatch, None]:

        remembered_catch_state = parent_match.remembered_catch_state
        if not remembered_catch_state:
            remembered_catch_state = self._create_on_catch_state(parent_match)
            if not remembered_catch_state:
                return None
            traverser.remember_on_catch(parent_match, remembered_catch_state)

            match = ImaginaryMatch(
                parent_match,
                "ImaginaryMatch",
                parent_match.data,
                self,
                vertex_index,
                parent_match.remembered_on_catch_match,
                parent_match.remembered_on_catch_action
            )

            return match

        try:
            item = next(remembered_catch_state.iterable)

            match = remembered_catch_state.match_constructor(
                parent_match,
                item[0],
                item[1],
                self,
                vertex_index - 1,  # reuse vertex on next get_match
                parent_match.remembered_on_catch_match,
                parent_match.remembered_on_catch_action
            )

            child_recursive_match = self.match(match, traverser, vertex_index)
            if child_recursive_match:
                return child_recursive_match
            return match

        except StopIteration:
            traverser.restore_on_catch(parent_match)
            return None

    def _create_on_catch_state(self, parent_match: TraverserMatch):
        data = parent_match.data
        if isinstance(data, dict):
            return self._key_match(parent_match)
        elif isinstance(data, list):
            return self._list_match(parent_match)
        else:
            return None

    @staticmethod
    def _key_match(parent_match: TraverserMatch) -> CatchState:
        iterable = iter(parent_match.data.items())
        return CatchState(iterable, KeyMatch)

    @staticmethod
    def _list_match(parent_match: TraverserMatch) -> CatchState:
        iterable = iter(enumerate(parent_match.data))
        return CatchState(iterable, ListMatch)



from treepath.path.builder.symbol import Symbol

# wildcard path constants.  When included in the path, the traversal algorithm applies the path to each sibling
# tree node.  It stops each time a successful match occurs.
# Examples:   path.wildcard, path[wildcard] path.a.wildcard, path.a[wildcard], path.wildcard.b, path[wildcard].b
# path.a.wildcard.b, path.a[wildcard].b
wildcard = Symbol("wildcard")

# wc is short version of wildcard
# Examples:   path.wc, path[wc] path.a.wc, path.a[wc], path.wc.b, path[wc].b
# path.a.wc.b, path.a[wc].b
wc = wildcard

# recursive path constants.  When included in the path, the traversal algorithm applies the path to each node in the
# in the tree in pre-order traversal.  It stops each time a successful match occurs.
# Examples:   path.recursive, path.a.recursive,  path.recursive.b, path.a.recursive.b
recursive = Symbol("recursive")

# rec is short version of recursive
# Examples:   path.recursive, path.a.recursive,  path.recursive.b, path.a.recursive.b
rec = recursive



from typing import Union, Iterator, Callable, Any, Tuple

from treepath.path.builder.path_builder import PathBuilder, get_vertex_from_path_builder
from treepath.path.builder.path_builder_predicate import PathBuilderPredicate
from treepath.path.builder.path_predicate import PathPredicate
from treepath.path.exceptions.match_not_found_error import MatchNotFoundError
from treepath.path.exceptions.nested_match_not_found_error import NestedMatchNotFoundError
from treepath.path.traverser.has_function import create_has_predicate
from treepath.path.traverser.match import Match
from treepath.path.traverser.match_traverser import MatchTraverser
from treepath.path.traverser.nested_match_traverser import NestedMatchTraverser
from treepath.path.traverser.nested_value_traverser import NestedValueTraverser
from treepath.path.traverser.predicate_match import PredicateMatch
from treepath.path.traverser.trace import Trace
from treepath.path.traverser.value_traverser import ValueTraverser
from treepath.path.util.decorator import pretty_repr, add_attr

_has_typing_first_arg = Union[
    PathBuilderPredicate,
    PathPredicate,
    Callable[[Match], Any]
]
_has_typing_single_arg_functions = Callable[[Any], Any]

_has_tuple_arg_type = Tuple[
    _has_typing_first_arg,
    _has_typing_single_arg_functions,
]

_has_multiple_arg_type = Union[
    _has_typing_first_arg,
    _has_tuple_arg_type,
]

# The value used to indicate an argument is not set.
_not_set = dict()


def get(
        expression: PathBuilder,
        data: Union[dict, list, Match],
        default=_not_set,
        trace: Callable[[Trace], None] = None
) -> Union[dict, list, str, int, float, bool, None]:
    """
    Returns the first value in the data that satisfies the path expression.   When no result is found, a
    MatchNotFoundError is raised unless a default value is given, In which case the default value is returned.

    @param expression: The path expression that define the search criteria.
    @param data: The data to search through.  The data must be either a tree structure that adheres to
        https://docs.python.org/3/library/json.html or a Match object from a previous search.
    @param default:  An optional value to return when no result is found.
    @param trace: An optional callable to report detail iteration data too.
    @return: The value that satisfies the path expression, else MatchNotFoundError is raised unless default is given.
    @raise MatchNotFoundError:  Raised when no result is found and no default value is given.
    """
    must_match = (default is _not_set)
    match = get_match(expression, data, must_match=must_match, trace=trace)
    if match:
        return match.data
    return default


def find(
        expression: PathBuilder,
        data: Union[dict, list, Match],
        trace: Callable[[Trace], None] = None
) -> Iterator[Union[dict, list, str, int, float, bool, None]]:
    """
    Construct a lazy iterator of all values in the data that satisfies the path expression.

    @param expression: The path expression that define the search criteria.
    @param data: The data to search through.  The data must be either a tree structure that adheres to
        https://docs.python.org/3/library/json.html or a Match object from a previous search.
    @param trace: An optional callable to report detail iteration data too.
    @return: A lazy iterator containing all values that satisfies the path expression.
    """
    if isinstance(data, Match):
        return nested_find(expression, data, trace=trace)

    vertex = get_vertex_from_path_builder(expression)
    traverser = ValueTraverser(data, vertex, trace=trace)
    traverser_iter = iter(traverser)
    return traverser_iter


def get_match(
        expression: PathBuilder,
        data: Union[dict, list, Match],
        must_match: bool = True,
        trace: Callable[[Trace], None] = None
) -> Union[Match, None]:
    """
    Returns the first Match in the data that satisfies the path expression.   When no result is found, a
    MatchNotFoundError is raised unless must_match is False, In which case None is returned.

    @param expression: The path expression that define the search criteria.
    @param data: The data to search through.  The data must be either a tree structure that adheres to
        https://docs.python.org/3/library/json.html or a Match object from a previous search.
    @param must_match:  An optional argument to indicate whether to raise MatchNotFoundError or return None when no
        result is found. By default MatchNotFoundError will be raise.
    @param trace: An optional callable to report detail iteration data too.
    @return: The Match that satisfies the path expression, else MatchNotFoundError is raised unless default is given.
    @raise MatchNotFoundError:  Raised when no result is found and must_match is set to True.
    """
    if isinstance(data, Match):
        return nested_get_match(expression, data, must_match=must_match, trace=trace)

    vertex = get_vertex_from_path_builder(expression)
    traverser = MatchTraverser(data, vertex, trace=trace)
    traverser_iter = iter(traverser)
    try:
        return next(traverser_iter)
    except StopIteration:
        pass
    if must_match:
        raise MatchNotFoundError(vertex)
    return None


def find_matches(
        expression: PathBuilder,
        data: Union[dict, list, Match],
        trace: Callable[[Trace], None] = None
) -> Iterator[Match]:
    """
    Construct a lazy iterator of all Matches in the data that satisfies the path expression.

    @param expression: The path expression that define the search criteria.
    @param data: The data to search through.  The data must be either a tree structure that adheres to
        https://docs.python.org/3/library/json.html or a Match object from a previous search.
    @param trace: An optional callable to report detail iteration data too.
    @return: A lazy iterator containing all Matches that satisfies the path expression.
    """
    if isinstance(data, Match):
        return nested_find_matches(expression, data, trace=trace)

    vertex = get_vertex_from_path_builder(expression)
    traverser = MatchTraverser(data, vertex, trace=trace)
    traverser_iter = iter(traverser)
    return traverser_iter


def nested_find(
        expression: PathBuilder,
        parent_match: Match,
        trace: Callable[[Trace], None] = None
) -> Iterator[Union[dict, list, str, int, float, bool, None]]:
    """
    Construct a lazy iterator of all values in the parent Match that satisfies the path expression. As a convenience,
    the find function also accepts the Match object as source data, so it not necessary to use this function directly.

    @param expression: The path expression that define the search criteria.
    @param parent_match: The data to search through.  The data must be a Match object from a previous search.
    @param trace: An optional callable to report detail iteration data too.
    @return: A lazy iterator containing all values that satisfies the path expression.
    """
    vertex = get_vertex_from_path_builder(expression)
    traverser = NestedValueTraverser(parent_match._traverser_match, vertex, trace=trace)
    traverser_iter = iter(traverser)
    return traverser_iter


def nested_get_match(
        expression: PathBuilder,
        parent_match: Match,
        must_match: bool = True,
        trace: Callable[[Trace], None] = None
) -> Union[Match, None]:
    """
    Returns the first Match in the parent Match that satisfies the path expression.   When no result is found, a
    MatchNotFoundError is raised unless must_match is False, In which case None is returned.  As a convenience, the
    get_match function also accepts the Match object as source data, so it not necessary to use this function directly.

    @param expression: The path expression that define the search criteria.
    @param parent_match: The data to search through.  The data must be a Match object from a previous search.
    @param must_match:  An optional argument to indicate whether to raise MatchNotFoundError or return None when no
        result is found. By default MatchNotFoundError will be raise.
    @param trace: An optional callable to report detail iteration data too.
    @return: The Match that satisfies the path expression, else MatchNotFoundError is raised unless default is given.
    @raise MatchNotFoundError:  Raised when no result is found and must_match is set to True.
    """
    if isinstance(parent_match, PredicateMatch):
        trace = parent_match.trace

    vertex = get_vertex_from_path_builder(expression)
    traverser = NestedMatchTraverser(parent_match._traverser_match, vertex, trace=trace)
    traverser_iter = iter(traverser)
    try:
        return next(traverser_iter)
    except StopIteration:
        pass
    if must_match:
        raise NestedMatchNotFoundError(parent_match, vertex)
    return None


def nested_find_matches(
        expression: PathBuilder,
        parent_match: Match,
        trace: Callable[[Trace], None] = None
) -> Iterator[Match]:
    """
    Construct a lazy iterator of all Matches in the parent Match that satisfies the path expression.  As a convenience,
    the find_matches function also accepts the Match object as source data, so it not necessary to use this function
    directly.

    @param expression: The path expression that define the search criteria.
    @param parent_match: The data to search through.  The data must be a Match object from a previous search.
    @param trace: An optional callable to report detail iteration data too.
    @return: A lazy iterator containing all Matches that satisfies the path expression.
    """
    if isinstance(parent_match, PredicateMatch):
        trace = parent_match.trace
    vertex = get_vertex_from_path_builder(expression)
    traverser = NestedMatchTraverser(parent_match._traverser_match, vertex, trace=trace)
    traverser_iter = iter(traverser)
    return traverser_iter


def has_these(*args: _has_multiple_arg_type, repr_join_key=', '):
    """
    The has these decorator defines an aggregates predicate.   It augments the path predicate its decorating with the
    decorator arguments.

    For example:
    #          arg1     arg2       arg3                      arg4
    @has.these(path.a,  path.b==2, (path.c, operator.truth), other_predicate)
    def predicate(match: Match, arg1, arg2, arg3, arg4):
     return return arg1(match) and arg2(match) and arg3(match) and arg4(match)
    value = get(path.rec[predicate].name,solar_system)

    @param args: Varying number of positional arguments.  Each argument must be a path expression or a path predicate.
    @param repr_join_key:  A string to use as a delimiter when joining the arguments while pretty printing path.
    @return: The decorated function to be used as a path predicate in a path expression.
    """

    def process_has_arg(arg):
        if isinstance(arg, tuple):
            return create_has_predicate(nested_find_matches, *arg)
        else:
            return create_has_predicate(nested_find_matches, arg)

    has_predicates = [process_has_arg(arg) for arg in args]

    def wrap(function):
        @pretty_repr(lambda: f"{repr_join_key.join(map(repr, has_predicates))}")
        def predicate(parent_match):
            return function(parent_match, *has_predicates)

        return predicate

    return wrap


def has_all(*args: _has_multiple_arg_type) -> Callable[[Match], Any]:
    """
    Construct a logical and predicate.   The outcome of evaluating the predicate is equivalent to
    return arg0(match) and arg1(match) and arg2(match) ... and argN(match)

    Example Usage
    * get(path.a[has_all(path.b, path.z)].c) interpret as get a.c if both a.b and a.z exist.
    * get(path.a[has_all(path.b, path.z==1)].c) interpret as get a.c if a.b exist and a.z==1.
    * get(path.a[has_all(path.b, (path.z==1, int))].c) interpret as get a.c if a.b exist and int(a.z)==1.
    * get(path.a[has_all(path.b, has(path.z==1, int))].c) has same meaning as previous.
    * get(path.a[has_all(has_any(path.l,path.m),has_any(path.x,path.y))].c) interpret as get a.c if either a.l or a.m
      exist and if either a.x or a.y exist.

    @param args: A variable length argument where each argument may be a path expression,  a path expression with
           condition operator, a predicate or a tuple.  A tuple argument is interpreted as a has function.
    @return:  Returns a predicate that performs a logical and on the arguments.
    """

    @has.these(*args, repr_join_key=' and ')
    def and_predicate(parent_match: Match, *predicates) -> Any:
        for predicate in predicates:
            if not predicate(parent_match):
                return False
        return True

    return and_predicate


def has_any(*args: _has_multiple_arg_type) -> Callable[[Match], Any]:
    """
    Construct a logical or predicate.   The outcome of evaluating the predicate is equivalent to
    return arg0(match) or arg1(match) or arg2(match) ... or argN(match)

    Example Usage
    * get(path.a[has_any(path.b, path.z)].c) interpret as get a.c if either a.b or a.z exist.
    * get(path.a[has_any(path.b, path.z==1)].c) interpret as get a.c if either a.b exist or a.z==1.
    * get(path.a[has_any(path.b, (path.z==1, int))].c) interpret as get a.c if either a.b exist or int(a.z)==1.
    * get(path.a[has_any(path.b, has(path.z==1, int))].c) has same meaning as previous.
    * get(path.a[has_any(has_all(path.l,path.m),has_all(path.x,path.y))].c) interpret as get a.c if a.l and a.m
      exist or if a.x and a.y exist.

    @param args: A variable length argument where each argument may be a path expression,  a path expression with
           condition operator, a predicate or a tuple.  A tuple argument is interpreted as a has function.
    @return:  Returns a predicate that performs a logical or on the arguments.
    """

    @has.these(*args, repr_join_key=' or ')
    def or_predicate(parent_match: Match, *predicates) -> Any:
        for predicate in predicates:
            if predicate(parent_match):
                return True
        return False

    return or_predicate


def has_not(
        path: _has_typing_first_arg,
        *single_arg_functions: _has_typing_single_arg_functions) -> Callable[[Match], Any]:
    """
    Construct a logical not predicate.  The has_not functions accepts the same
    arguments as the has function.

    Example Usage
     * get(path.a[has_not(path.b)].c) interpret as get a.c if  a.b does not exist.
     * get(path.a[has_not(has(path.b))].c) has same meaning as previous.
     * get(path.a[has_not(has_any(path.b, path.z))].c) interpret as get a.c if both a.b and a.z do not exist.


    @param path: A path expression,  a path expression with condition operator, or predicate.
    @param single_arg_functions:  A variable argument of or Callable[[ANY],ANY].   The Callables are evaluated as
           follows: arg0(arg1(arg2(...argN(get(path)))).

    @return:  Returns the constructed not predicate.
    """

    predicate = create_has_predicate(nested_find_matches, path, *single_arg_functions)

    def not_predicate(parent_match: Match) -> Any:
        return not predicate(parent_match)

    return not_predicate


@add_attr("these", has_these)
def has(
        path: _has_typing_first_arg,
        *single_arg_functions: _has_typing_single_arg_functions) -> Callable[[Match], Any]:
    """
    Constructs a predicate from a path expression, a path expression and condition, a path expression, a
    condition and a map function or a predicate.  These four forms are shown here:

    First form: has(path expression)
    * Evaluates true if the path expression combined with parent path expression exist.
    * get(path.a[has(path.b)].c) interpret as get a.c if a.b exist.

    Second form: has(path expression, condition)
    * Applies the condition to the value referenced by the path expression combined with parent path expression.
    * The condition can be either a single argument function that returns a value or any of the following operators:
      ==, !=, >, >=, <, or <=.
    * get(path.a[has(path.b, operator.truth)].c) interpret as get a.c if operator.truth(get(a.b)).
    * get(path.a[has(path.b==1)].c) interpret as get a.c if get(a.b)==1.

    Third form: has(path expression, condition, map function)
    * First applies the map function to the value referenced by the path expression combined with parent path
      expression. Next apply the condition to the value return from map function.
    * The map function must be a single argument function that returns a value.
    * get(path.a[has(path.b,  operator.truth, int)].c) interpret as get a.c if operator.truth(int(get(a.b))).
    * get(path.a[has(path.b==1, int)].c) interpret as get a.c if int(get(a.b))==1.

    Fourth form: has(path predicate)
    * Return the path predicate argument.
    * A path predicate type:  Callable[[Match], ANY]
    * get(path.a[has(has(path.b))].c) == get(path.a[has(path.b)].c)

    @param path: A path expression,  a path expression with condition operator, or predicate.
    @param single_arg_functions:  A variable argument of or Callable[[ANY],ANY].   The Callables are evaluated as
           follows: arg0(arg1(arg2(...argN(get(path)))).

    @return:  Returns the constructed predicate.
    """
    return create_has_predicate(nested_find_matches, path, *single_arg_functions)



from typing import Union

from treepath.path.traverser.key_match import KeyMatch
from treepath.path.traverser.traverser_match import TraverserMatch
from treepath.path.vertex.vertex import Vertex


class KeyVertex(Vertex):
    __slots__ = ()

    def __init__(self, parent, name):
        super().__init__(parent, name)

    @property
    def path_segment(self):
        return f".{self.name}"

    def match(self, parent_match: TraverserMatch, traverser, vertex_index: int) -> Union[TraverserMatch, None]:
        data = parent_match.data
        if not isinstance(data, dict):
            return None
        try:
            name = self.name
            value = data[self.name]
            return KeyMatch(
                parent_match,
                name,
                value,
                self,
                vertex_index,
                parent_match.remembered_on_catch_match,
                parent_match.remembered_on_catch_action
            )
        except KeyError:
            pass


class KeyWildVertex(Vertex):
    __slots__ = ()

    def __init__(self, parent):
        super().__init__(parent, '*')
        self.is_catch_vertex = True

    @property
    def path_segment(self):
        return ".*"

    def match(self, parent_match: TraverserMatch, traverser, vertex_index: int) -> Union[TraverserMatch, None]:

        remembered_catch_state = parent_match.remembered_catch_state

        # create a dict iterator  if there is none
        if not remembered_catch_state:
            data = parent_match.data
            if not isinstance(data, dict):
                return None
            remembered_catch_state = iter(data.items())
            traverser.remember_on_catch(parent_match, remembered_catch_state)

        try:
            item = next(remembered_catch_state)
            return KeyMatch(
                parent_match,
                item[0],
                item[1],
                self,
                vertex_index,
                parent_match.remembered_on_catch_match,
                parent_match.remembered_on_catch_action
            )
        except StopIteration:
            traverser.restore_on_catch(parent_match)
            return None



import functools
from typing import Union, Callable, Any

from treepath.path.builder.path_builder_predicate import PathBuilderPredicate
from treepath.path.builder.path_predicate import PathPredicate
from treepath.path.exceptions.path_syntax_error import PathSyntaxError
from treepath.path.traverser.match import Match
from treepath.path.util.decorator import pretty_repr


def create_has_predicate(
        nested_find_matches,
        path: Union[PathBuilderPredicate, PathPredicate,  Callable[[Match], Any]],
        *value_remap: [Callable[[Any], Any]]) -> Callable[[Match], Any]:
    if isinstance(path, PathPredicate):
        real_path = path.path
        single_arg_operator = path.operation
    elif isinstance(path, PathBuilderPredicate):
        real_path = path
        single_arg_operator = None
    elif isinstance(path, Callable):
        return path
    else:
        raise PathSyntaxError(
            parent_vertex=None,
            error_msg=f"Invalid  path [{type(path)}] argument.   Expecting PathBuilderPredicate, PathPredicate,  or "
                      'Callable[[Match], Any]]',
            invalid_path_segment='')

    match_iter = functools.partial(nested_find_matches, real_path)

    if single_arg_operator and value_remap:
        return _create_has_predicate_with_single_arg_operator_and_value_remap(
            real_path,
            match_iter,
            single_arg_operator,
            value_remap
        )

    if not single_arg_operator and value_remap:
        return _create_has_predicate_with_value_remap(real_path, match_iter, value_remap)

    if single_arg_operator and not value_remap:
        return _create_has_predicate_with_single_arg_operator(real_path, match_iter, single_arg_operator)
    else:
        return _create_has_predicate(real_path, match_iter)


def _create_has_predicate_with_single_arg_operator_and_value_remap(
        real_path,
        match_iter,
        single_arg_operator,
        value_remap
):
    @pretty_repr(
        lambda: f"has({real_path} {single_arg_operator}, {', '.join(map(repr, value_remap))})")
    def has_predicate(parent_match: Match):
        for next_match in match_iter(parent_match):

            value = next_match.data
            for function in value_remap[::-1]:
                value = function(value)

            value = single_arg_operator(value)

            if value:
                return True
        return False

    return has_predicate


def _create_has_predicate_with_value_remap(real_path, match_iter, value_remap):
    @pretty_repr(lambda: f"has({real_path}, {', '.join(map(repr, value_remap))})")
    def has_predicate(parent_match: Match):
        for next_match in match_iter(parent_match):

            value = next_match.data
            for function in value_remap[::-1]:
                value = function(value)

            if value:
                return True
        return False

    return has_predicate


def _create_has_predicate_with_single_arg_operator(real_path, match_iter, single_arg_operator):
    @pretty_repr(lambda: f"has({real_path} {single_arg_operator})")
    def has_predicate(parent_match: Match):
        for next_match in match_iter(parent_match):
            value = single_arg_operator(next_match.data)
            if value:
                return True
        return False

    return has_predicate


def _create_has_predicate(real_path, match_iter):
    @pretty_repr(lambda: f"has({real_path})")
    def has_predicate(parent_match: Match):
        for _ in match_iter(parent_match):
            return True
        return False

    return has_predicate



from __future__ import annotations

from typing import Union, Callable, Match, Any

from treepath.path.builder.abstract_class_builder import AbstractPathBuilder
from treepath.path.builder.patch_constants import wildcard
from treepath.path.builder.path_builder_predicate import PathBuilderPredicate
from treepath.path.builder.symbol import Symbol
from treepath.path.exceptions.path_syntax_error import PathSyntaxError
from treepath.path.vertex.key_vertex import KeyVertex, KeyWildVertex
from treepath.path.vertex.list_vertex import ListIndexVertex, ListSliceVertex, ListWildVertex
from treepath.path.vertex.parent_vertex import ParentVertex
from treepath.path.vertex.predicate_vertex import PredicateVertex
from treepath.path.vertex.recursive_vertex import RecursiveVertex
from treepath.path.vertex.tuple_vertex import TupleVertex
from treepath.path.vertex.vertex import Vertex

_RESERVED_ATTR_FOR_VERTEX_DATA = "_RESERVED_ATTR_FOR_VERTEX_DATA"
"""
  _RESERVED_ATTR_FOR_VERTEX_DATA:  A special attribute name used by PathBuilder to store all the DATA related to the 
  vertex.   It purpose keep data attribute separate from PathBuilder in order to reduce attribute name collision.  
"""


def _build_key(parent_vertex: Vertex, key: Union[int, slice, Symbol, str, tuple, Callable[[Match], Any]]):
    if isinstance(key, int):
        return ListIndexVertex(parent_vertex, key)
    elif isinstance(key, slice):
        return ListSliceVertex(parent_vertex, key)
    if wildcard == key:
        return ListWildVertex(parent_vertex)
    elif isinstance(key, str):
        return KeyVertex(parent_vertex, key)
    elif isinstance(key, tuple):
        return TupleVertex(parent_vertex, key)
    elif callable(key):
        return PredicateVertex(parent_vertex, key)
    else:
        raise PathSyntaxError(parent_vertex,
                              f"Unsupported indices: [{type(key)}].  Indices must be either an int, slice, tuple, "
                              f"str, wildcard, has(function), or callable.",
                              f"[{key}]"
                              )


class PathBuilder(PathBuilderPredicate, AbstractPathBuilder):
    __slots__ = _RESERVED_ATTR_FOR_VERTEX_DATA

    def __init__(self, vertex, ):
        object.__setattr__(self, _RESERVED_ATTR_FOR_VERTEX_DATA, vertex)

    def __getattr__(self, name):
        """
        creates a new attribute at self.
        This method only get called if self.name does not already exist
        self.name
        """
        parent_vertex = object.__getattribute__(self, _RESERVED_ATTR_FOR_VERTEX_DATA)
        vertex = KeyVertex(parent_vertex, self.tranform_attribute_name(name))
        path_builder = self.create_path_builder(vertex)
        return path_builder

    def __setattr__(self, name, value):
        """
        Overloads self.name = value
        Attribute assignment not supported.
        """
        raise AttributeError(f"set __setattr__  {name} {value} not supported on {self.__name__}")

    def __getitem__(self, key: Union[int, slice, Symbol, str, tuple, Callable[[Match], Any]]):
        """
        Overloads self[key]
        """
        parent_vertex = object.__getattribute__(self, _RESERVED_ATTR_FOR_VERTEX_DATA)
        vertex = _build_key(parent_vertex, key)
        path_builder = self.create_path_builder(vertex)
        return path_builder

    def __setitem__(self, key, value):
        """
        Overloads self[key]=value
        """
        raise AttributeError(f"set __setitem__  {key} not supported on {self.__name__}")

    def __repr__(self):
        vertex = object.__getattribute__(self, _RESERVED_ATTR_FOR_VERTEX_DATA)
        return repr(vertex)

    def __str__(self):
        vertex = object.__getattribute__(self, _RESERVED_ATTR_FOR_VERTEX_DATA)
        return str(vertex)

    def __len__(self):
        """ Always return 0 as the [] is used as a factory and not a fetch"""
        return 0

    @property
    def shape(self):
        """
        The attribute name shape is reserved. Use path["shape"] instead of path.shape',

        Debuggers expect a property named shape to be defined when a class also defines a __getitem__.   The dynamite
        attribute defined by __getattr__ conflicts with the value the debugger expects the shape property to return.
        This can causes instability with some debuggers.

        To prevent any instability with debuggers, the property shape is reserved so it cannot be used a a dynamic
        property.
        """
        parent_vertex = object.__getattribute__(self, _RESERVED_ATTR_FOR_VERTEX_DATA)
        raise PathSyntaxError(
            parent_vertex,
            'The attribute name shape is reserved. Use path["shape"] instead of path.shape.',
            ".shape"
        )

    def create_path_builder(self, *args, **kwargs):
        return PathBuilder(*args, **kwargs)

    def tranform_attribute_name(self, name):
        return name

    @property
    def recursive(self):
        parent_vertex = object.__getattribute__(self, _RESERVED_ATTR_FOR_VERTEX_DATA)
        vertex = RecursiveVertex(parent_vertex)
        if isinstance(parent_vertex, RecursiveVertex):
            raise PathSyntaxError(
                parent_vertex,
                "Successive recursive vertices are not allowed in the path expression.",
                ".."
            )
        path_builder = self.create_path_builder(vertex)
        return path_builder

    @property
    def rec(self):
        return self.recursive

    @property
    def wildcard(self):
        parent_vertex = object.__getattribute__(self, _RESERVED_ATTR_FOR_VERTEX_DATA)
        vertex = KeyWildVertex(parent_vertex)
        path_builder = self.create_path_builder(vertex)
        return path_builder

    @property
    def wc(self):
        return self.wildcard

    @property
    def parent(self):
        parent_vertex = object.__getattribute__(self, _RESERVED_ATTR_FOR_VERTEX_DATA)
        vertex = ParentVertex(parent_vertex)
        path_builder = self.create_path_builder(vertex)
        return path_builder


def get_vertex_from_path_builder(path_builder: PathBuilder) -> Vertex:
    vertex = object.__getattribute__(path_builder, _RESERVED_ATTR_FOR_VERTEX_DATA)
    return vertex



import operator

from treepath.path.builder.path_predicate import PathPredicate
from treepath.path.util.decorator import pretty_repr
from treepath.path.util.function import create_partial_operator


class PathBuilderPredicate:
    __slots__ = ()

    def __lt__(self, other) -> PathPredicate:
        operator_ = create_partial_operator(operator.__lt__, other)
        pretty = pretty_repr(lambda: f"< {other}")
        return PathPredicate(self, pretty(operator_))

    def __le__(self, other) -> PathPredicate:
        operator_ = create_partial_operator(operator.__le__, other)
        pretty = pretty_repr(lambda: f"<= {other}")
        return PathPredicate(self, pretty(operator_))

    def __eq__(self, other) -> PathPredicate:
        operator_ = create_partial_operator(operator.__eq__, other)
        pretty = pretty_repr(lambda: f"== {other}")
        return PathPredicate(self, pretty(operator_))

    def __ne__(self, other) -> PathPredicate:
        operator_ = create_partial_operator(operator.__ne__, other)
        pretty = pretty_repr(lambda: f"!= {other}")
        return PathPredicate(self, pretty(operator_))

    def __gt__(self, other) -> PathPredicate:
        operator_ = create_partial_operator(operator.__gt__, other)
        pretty = pretty_repr(lambda: f"> {other}")
        return PathPredicate(self, pretty(operator_))

    def __ge__(self, other) -> PathPredicate:
        operator_ = create_partial_operator(operator.__ge__, other)
        pretty = pretty_repr(lambda: f">= {other}")
        return PathPredicate(self, pretty(operator_))



from treepath.path.exceptions.match_not_found_error import MatchNotFoundError
from treepath.path.traverser.match import Match
from treepath.path.vertex.vertex import Vertex


class NestedMatchNotFoundError(MatchNotFoundError, LookupError):
    """
    NestedMatchNotFoundError is raised when nested_get_match query fails to generate a result for the specified path.
    """

    def __init__(self, nested_match: Match, vertex: Vertex):
        super().__init__(vertex)
        self.nested_match = nested_match

    def _resolve_msg(self):
        path = repr(self.vertex)
        nested_path = self.nested_match.path
        return f"NestedMatchNotFoundError(No get_match occurred on path {path} from match {nested_path})"



from typing import Union

from treepath.path.traverser.nested_match_traverser import NestedMatchTraverser


class NestedValueTraverser(NestedMatchTraverser):
    __slots__ = ()

    def __next__(self) -> Union[dict, list, str, int, float, bool, None]:
        return super().__next__().data



from typing import Union

from treepath.path.traverser.parent_match import ParentMatch
from treepath.path.traverser.traverser_match import TraverserMatch
from treepath.path.vertex.vertex import Vertex


class ParentVertex(Vertex):
    __slots__ = ()

    def __init__(self, parent):
        super().__init__(parent, "parent")

    @property
    def path_segment(self):
        return f".{self.name}"

    def match(self, parent_match: TraverserMatch, traverser, vertex_index: int) -> Union[TraverserMatch, None]:
        remembered_parent = parent_match.remembered_parent

        if not remembered_parent:
            return None

        return ParentMatch(
            remembered_parent,
            parent_match,
            remembered_parent.data_name,
            remembered_parent.data,
            self,
            vertex_index,
            parent_match.remembered_on_catch_match,
            parent_match.remembered_on_catch_action
        )



from abc import abstractmethod
from typing import Union

from treepath.path.traverser.list_match import ListMatch
from treepath.path.traverser.traverser_match import TraverserMatch
from treepath.path.util.function import enumerate_slice
from treepath.path.vertex.vertex import Vertex


class _ListVertex(Vertex):
    __slots__ = ()

    def __init__(self, parent, name):
        super().__init__(parent, name)

    @property
    def path_segment(self):
        return f"[{self.name}]"

    @abstractmethod
    def match(self, parent_match: TraverserMatch, traverser, vertex_index: int) -> Union[TraverserMatch, None]:
        raise NotImplementedError


class ListIndexVertex(_ListVertex):
    __slots__ = 'index'

    def __init__(self, parent, index: int):
        self.index = index
        super().__init__(parent, index)

    def match(self, parent_match: TraverserMatch, traverser, vertex_index: int) -> Union[TraverserMatch, None]:
        data = parent_match.data
        if not isinstance(data, list):
            return None
        try:
            index = self.index
            value = data[index]
            return ListMatch(
                parent_match,
                index,
                value,
                self,
                vertex_index,
                parent_match.remembered_on_catch_match,
                parent_match.remembered_on_catch_action
            )
        except IndexError:
            pass


class ListSliceVertex(_ListVertex):
    __slots__ = '_slice'

    def __init__(self, parent, slice_: slice):
        self._slice = slice_
        super().__init__(parent, slice_)
        self.is_catch_vertex = True

    @property
    def path_segment(self):
        slice_ = self._slice
        start = slice_.start
        if not start:
            start = ''
        stop = slice_.stop
        if not stop:
            stop = ''
        step = slice_.step
        if not step:
            return f"[{start}:{stop}]"
        return f"[{start}:{stop}:{step}]"

    def match(self, parent_match: TraverserMatch, traverser, vertex_index: int) -> Union[TraverserMatch, None]:

        remembered_catch_state = parent_match.remembered_catch_state

        # create a slice iterator  if there is none
        if not remembered_catch_state:
            data = parent_match.data
            if not isinstance(data, list):
                return None
            remembered_catch_state = enumerate_slice(self._slice, data)
            traverser.remember_on_catch(parent_match, remembered_catch_state)

        try:
            item = next(remembered_catch_state)
            return ListMatch(
                parent_match,
                item[0],
                item[1],
                self,
                vertex_index,
                parent_match.remembered_on_catch_match,
                parent_match.remembered_on_catch_action
            )
        except StopIteration:
            traverser.restore_on_catch(parent_match)
            return None


class ListWildVertex(_ListVertex):
    __slots__ = ()

    def __init__(self, parent):
        super().__init__(parent, '*')
        self.is_catch_vertex = True

    @property
    def path_segment(self):
        return "[*]"

    def match(self, parent_match: TraverserMatch, traverser, vertex_index: int) -> Union[TraverserMatch, None]:

        remembered_catch_state = parent_match.remembered_catch_state

        # create a list iterator  if there is none
        if not remembered_catch_state:
            data = parent_match.data
            if not isinstance(data, list):
                return None
            remembered_catch_state = enumerate(data)
            traverser.remember_on_catch(parent_match, remembered_catch_state)

        try:
            item = next(remembered_catch_state)
            return ListMatch(
                parent_match,
                item[0],
                item[1],
                self,
                vertex_index,
                parent_match.remembered_on_catch_match,
                parent_match.remembered_on_catch_action
            )
        except StopIteration:
            traverser.restore_on_catch(parent_match)
            return None



from typing import List


def create_partial_operator(operator_, right):
    def partial_operator_(left):
        return operator_(left, right)

    return partial_operator_


def enumerate_slice(_slice: slice, _list: List):
    indices = _slice.indices(len(_list))
    slice_range = iter(range(*indices))

    def slice_enumerate(item):
        return next(slice_range), item

    return map(slice_enumerate, _list[_slice])
